@@ -20,8 +20,12 @@ from utils import comma_list_sanitize
 from version import Version
 from version import VersionFilter
 
-BUILDSLAVE_KWARGS = ('max_builds',)
-BUILDSLAVE_REQUIRED = ('password', 'pg_version',)
+BUILDSLAVE_KWARGS = { # name -> validating callable
+    'max_builds' : int,
+    'notify_on_missing': str,
+    }
+
+BUILDSLAVE_REQUIRED = ('password',)
 
 logger = logging.getLogger(__name__)
 
@@ -75,13 +79,29 @@ class BuildoutsConfigurator(object):
         for slavename in parser.sections():
             kw = {}
             kw['properties'] = props = {}
+            props['capability'] = caps = {} # name -> versions
             seen = set()
             for key, value in parser.items(slavename):
                 seen.add(key)
                 if key in ('passwd', 'password'):
                     pwd = value
+                elif key == 'capability':
+                    for cap_line in value.split(os.linesep):
+                        split = cap_line.split()
+                        if len(split) == 1:
+                            name = split[0]
+                            version = None
+                        else:
+                            name = split[0]
+                            version = split[1]
+                        this_cap = caps.setdefault(name, {})
+                        cap_opts = this_cap.setdefault(version, {})
+                        for option in split[2:]:
+                            opt_name, opt_val = option.split('=')
+                            cap_opts[opt_name] = opt_val
+
                 elif key in BUILDSLAVE_KWARGS:
-                    kw[key] = value
+                    kw[key] = BUILDSLAVE_KWARGS[key](value)
                 else:
                     props[key] = value
 
@@ -91,14 +111,64 @@ class BuildoutsConfigurator(object):
                                  slavename, option)
                     break
             else:
-                slaves.append(BuildSlave(slavename, pwd, **kw))
+                slave = BuildSlave(slavename, pwd, **kw)
+                slaves.append(slave)
 
         return slaves
 
-    def make_factory(self, name, cfg_path, options):
+    def buildout_standalone_dl_steps(self, cfg_tokens):
+        """Return slave side path and steps about the buildout.
+
+        The first returned value is the expected path from build directory
+        The second is an iterable of steps to get the buildout config file
+        and the related needed files (extended cfgs, bootstrap.py).
+        """
+        if len(cfg_tokens) != 1:
+            raise ValueError(
+                "Wrong standalong buildout specification: %r" % tokens)
+
+        conf_path = cfg_tokens[0]
+        conf_name = os.path.split(conf_path)[-1]
+        return conf_name, (FileDownload(mastersrc='buildouts/bootstrap.py',
+                                        slavedest='bootstrap.py'),
+                           FileDownload(mastersrc=conf_path,
+                                        slavedest=conf_name),
+                           )
+
+    def buildout_hg_dl_steps(self, cfg_tokens):
+        """Return slave side path and steps about the buildout.
+
+        The first returned value is the expected path from build directory
+        The second is an iterable of steps to get the buildout config file
+        and the related needed files (extended cfgs, bootstrap.py).
+        """
+        if len(cfg_tokens) != 3:
+            raise ValueError(
+                "Wrong standalong buildout specification: %r" % tokens)
+
+        url, branch, conf_path = cfg_tokens
+        return conf_path, (
+            FileDownload(
+                mastersrc='build_utils/buildout_hg_dl.py',
+                slavedest='buildout_hg_dl.py',
+                haltOnFailure=True),
+            ShellCommand(
+                command=['python', 'buildout_hg_dl.py', url, branch],
+                description=("Retrieve buildout", "from hg",),
+                haltOnFailure=True,
+                )
+            )
+
+    def make_factory(self, name, buildout_slave_path, buildout_dl_steps,
+                     options):
         """Return a build factory using name and buildout config at cfg_path.
 
-        cfg_path is relative to the master directory.
+        buildout_path is the slave-side path from build directory to the
+          buildout configuration file.
+        buildout_dl_steps is an iterable of BuildSteps to retrieve slave-side
+          the buildout configuration file and its dependencies (bootstrap,
+          extended conf files)
+
         the factory name is also used as testing database suffix
         options is the config part for this factory, seen as a dict
         """
@@ -108,15 +178,46 @@ class BuildoutsConfigurator(object):
                                      description="init bzr repo",
                                      flunkOnFailure=False,
                                      warnOnFailure=False,
+                                     hideStepIf=True,
                                      ))
-        factory.addStep(FileDownload(mastersrc='buildouts/bootstrap.py',
-                                     slavedest='bootstrap.py'))
-        factory.addStep(FileDownload(mastersrc=cfg_path,
-                                     slavedest='buildout.cfg'))
+
+        for dl_step in buildout_dl_steps:
+            factory.addStep(dl_step)
+
         factory.addStep(FileDownload(mastersrc='build_utils/'
                                      'analyze_oerp_tests.py',
                                      slavedest='analyze_oerp_tests.py'))
-        factory.addStep(ShellCommand(command=['python', 'bootstrap.py'],
+
+        def set_pg_cluster_props(step):
+            """Set postgresql cluster properties
+
+            All properties defined in the slave capability line for the
+            builder-level ``pg_version`` property are applied to the build,
+            with pg_prefix.
+
+            Example: port=5434 gives pg_port=5434
+
+            GR XXX Of course it is quite hacky to rely on side effect of a
+            doStepIf option. This is the quickest working solution I could come
+            with. Better to hook this in the build definition stage itself.
+            """
+            pg_version = step.getProperty('pg_version')
+            pg_props = step.getProperty('capability')['postgresql'][pg_version]
+            for k, v in pg_props.items():
+                step.setProperty('pg_%s' % k, v)
+            return True
+
+        factory.addStep(ShellCommand(
+            command=['/bin/echo',
+                     WithProperties('capability: %(capability)s')],
+            doStepIf=set_pg_cluster_props,
+            description=["Setting", "pg cluster", "properties"],
+            descriptionDone=["Set", "pg cluster", "properties"],
+            name="pg_cluster_props",
+            ))
+
+        factory.addStep(ShellCommand(command=['python', 'bootstrap.py',
+                                              '-c', buildout_slave_path],
                                      haltOnFailure=True,
                                      ))
 
@@ -134,6 +235,7 @@ class BuildoutsConfigurator(object):
                           )
         factory.addStep(ShellCommand(command=[
                     'bin/buildout',
+                    '-c', buildout_slave_path,
                     'buildout:eggs-directory=' + eggs_cache,
                     'buildout:openerp-downloads-directory=' + openerp_cache,
                     'openerp:vcs-clear-locks=True',
@@ -221,6 +323,13 @@ class BuildoutsConfigurator(object):
                 vf = VersionFilter.parse(line)
                 factory.build_for[vf.cap] = vf
 
+        requires = options.get('build-requires')
+        if requires is None:
+            factory.build_requires = []
+        else:
+            factory.build_requires = [r.strip()
+                                      for r in requires.split(os.linesep)]
+
         build_category = options.get('build-category')
         if build_category:
             factory.build_category = build_category.strip()
@@ -249,13 +358,17 @@ class BuildoutsConfigurator(object):
                 continue
 
             btype = buildout[0]
-            if btype != 'standalone':
+            buildout_downloader = self.buildout_dl_steps.get(btype)
+            if buildout_downloader is None:
                 raise ValueError("Buildout type %r in %r not supported" % (
                         btype, name))
 
-            registry[name] = self.make_factory(name, buildout[1],
+            conf_slave_path, dl_steps = buildout_downloader(self, buildout[1:])
+            registry[name] = self.make_factory(name, conf_slave_path, dl_steps,
                                                dict(parser.items(name)))
 
+    buildout_dl_steps = dict(standalone=buildout_standalone_dl_steps,
+                             hg=buildout_hg_dl_steps)
 
     def make_builders(self, master_config=None):
         """Spawn builders from build factories.
@@ -270,8 +383,6 @@ class BuildoutsConfigurator(object):
         The idea is notably to sort slaves according to capabilities (for
         specific requirements, such as ability to build a tricky python
         package) and environmental parameters (postgresql version etc.)
-
-        To demonstrate, for now, we iterate on pg_version
         """
 
         # this parameter is passed as kwarg for the sake of expliciteness
@@ -279,21 +390,31 @@ class BuildoutsConfigurator(object):
 
         slaves = master_config['slaves']
 
-        slaves_by_pg = {} # pg version -> list of slave names
+        slaves_by_pg = {} # pg version -> list of slaves
         for slave in slaves:
-            pg = slave.properties['pg_version']
-            slaves_by_pg.setdefault(pg, []).append(slave.slavename)
+            for pg in slave.properties['capability'].get('postgresql', {}).keys():
+                slaves_by_pg.setdefault(pg, []).append(slave)
 
         all_builders = []
         fact_to_builders = self.factories_to_builders
         for factory_name, factory in self.build_factories.items():
             pgvf = factory.build_for.get('postgresql')
+            requires = set(factory.build_requires)
+
+            meet_requires = {} # pg version -> list of slave names
+            for pg, slaves in slaves_by_pg.items():
+                meet = [slave.slavename for slave in slaves
+                        if requires.issubset(slave.properties['capability'])]
+                if meet:
+                    meet_requires[pg] = meet
+
             builders = [
                 BuilderConfig(name='%s-postgresql-%s' % (factory_name,
                                                          pg_version),
+                              properties=dict(pg_version=pg_version),
                               category=getattr(factory, 'build_category', None),
                               factory=factory, slavenames=slavenames)
-                for pg_version, slavenames in slaves_by_pg.items()
+                for pg_version, slavenames in meet_requires.items()
                 if pgvf is None or pgvf.match(Version.parse(pg_version))
                 ]
             fact_to_builders[factory_name] = [b.name for b in builders]
