@@ -156,7 +156,8 @@ def configure_logging(logging_level):
     logger = logging.getLogger()
     console_handler = logging.StreamHandler()
     console_handler.setFormatter(
-        logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s'))
+        logging.Formatter(
+            '%(asctime)-15s %(name)-12s %(levelname)-8s: %(message)s'))
     console_handler.setLevel(level)
 
     logger.addHandler(console_handler)
@@ -191,8 +192,8 @@ def update():
     try:
         fcntl.lockf(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
     except IOError:
-        print("Another instance of update-mirrors is running for buildmaster "
-              "%r . Exiting now." % options.buildmaster_dir)
+        print("Another instance of update-mirrors is running for mirrors "
+              "directory %r." % mirrors_dir)
         sys.exit(0)
 
     utils.vcs_binaries['bzr'] = options.bzr
